open Format
open Syntax
open Support.Error
open Support.Pervasive

(* ------------------------   EVALUATION  ------------------------ *)

exception NoRuleApplies

let rec isnumericval t = match t with
    TmZero(_) -> true
  | TmSucc(_,t1) -> isnumericval t1
  | _ -> false

let rec isval prog t = match t with
    TmTrue(_)  -> true
  | TmFalse(_) -> true
  | t when isnumericval t  -> true
  | TmFunApp(fi,fn,ts) -> 
      if hasbodyFun prog fn then false
      else List.for_all (isval prog) ts
  | _ -> false

let rec eval1 prog t = match t with
    TmIf(_,TmTrue(_),t2,t3) ->
      t2
  | TmIf(_,TmFalse(_),t2,t3) ->
      t3
  | TmIf(fi,t1,t2,t3) ->
      let t1' = eval1 prog t1 in
      TmIf(fi, t1', t2, t3)
  | TmSucc(fi,t1) ->
      let t1' = eval1 prog t1 in
      TmSucc(fi, t1')
  | TmPred(_,TmZero(_)) ->
      TmZero(dummyinfo)
  | TmPred(_,TmSucc(_,nv1)) when (isnumericval nv1) ->
      nv1
  | TmPred(fi,t1) ->
      let t1' = eval1 prog t1 in
      TmPred(fi, t1')
  | TmIsZero(_,TmZero(_)) ->
      TmTrue(dummyinfo)
  | TmIsZero(_,TmSucc(_,nv1)) when (isnumericval nv1) ->
      TmFalse(dummyinfo)
  | TmIsZero(fi,t1) ->
      let t1' = eval1 prog t1 in
      TmIsZero(fi, t1')
  | TmVar(fi,x) ->
      raise NoRuleApplies
  | TmFunApp(fi,fn,ts) ->
      let (vals,nonvals) = take_prefix (isval prog) ts in
      (match nonvals with
	[] -> 
	  let (argvs,bdy_opt) = getFunBodyFromProgram fi prog fn in
	  (match bdy_opt with
	    None -> raise NoRuleApplies (* stuck, no body for fn *)
	  | Some(bdy) -> simultSubstArgsVals fi argvs vals bdy)
      |	ti::rest -> 
	  let ti' = eval1 prog ti in
	  TmFunApp(fi,fn,vals@[ti']@rest))
  | _ -> 
      raise NoRuleApplies

let rec eval prog t =
  try let t' = eval1 prog t
      in eval prog t'
  with NoRuleApplies -> 
    if isval prog t then t
    else (warning "evaluation reached a stuck (non-value) term!";
	  t)

(* ------------------------   TYPING  ------------------------ *)

let rec typeof prog ctx t =
  match t with
    TmTrue(fi) -> 
      TyBool
  | TmFalse(fi) -> 
      TyBool
  | TmIf(fi,t1,t2,t3) ->
     if (=) (typeof prog ctx t1) TyBool then
       let tyT2 = typeof prog ctx t2 in
       if (=) tyT2 (typeof prog ctx t3) then tyT2
       else error fi "arms of conditional have different types"
     else error fi "guard of conditional not a boolean"
  | TmZero(fi) ->
      TyNat
  | TmSucc(fi,t1) ->
      if (=) (typeof prog ctx t1) TyNat then TyNat
      else error fi "argument of succ is not a number"
  | TmPred(fi,t1) ->
      if (=) (typeof prog ctx t1) TyNat then TyNat
      else error fi "argument of pred is not a number"
  | TmIsZero(fi,t1) ->
      if (=) (typeof prog ctx t1) TyNat then TyBool
      else error fi "argument of iszero is not a number"
  | TmVar(fi,x) -> getTypeFromContext fi ctx x
  | TmFunApp(fi,fn,ts) -> 
      let (argtys,resty) = getFunTypeFromProgram fi prog fn in
      let ttys = List.map (typeof prog ctx) ts in
      let rec checktyps tyAs tyBs = 
		match (tyAs,tyBs) with
		   ([],[]) -> resty 
		| (tyA::tyAs,tyB::tyBs) ->
		    if (=) tyA tyB then
		      checktyps tyAs tyBs
		    else
		      errfAt fi (fun ()->
			pr "Argument type mismatch in application:"; 
			pr " expected type "; printty tyA;
			pr " but got type "; printty tyB)
			
                | _ -> 
		    let msg =
		      Printf.sprintf 
			"Wrong number of arguments for function %s (expected %d, got %d)" in
		    error fi (msg fn (List.length argtys) (List.length ts)) in

      checktyps argtys ttys
