open Format
open Support.Error
open Support.Pervasive

(* ---------------------------------------------------------------------- *)
(* Datatypes *)

type ty =
    TyBool
  | TyNat

type term =
    TmTrue of info
  | TmFalse of info
  | TmIf of info * term * term * term
  | TmZero of info
  | TmSucc of info * term
  | TmPred of info * term
  | TmIsZero of info * term
  (* FOFL: *)
  | TmVar of info * string
  | TmFunApp of info * string * (term list)

(* context bindings *)

type binding =
    VarBind of ty

(* program definitions *)

type progdef = (* def _ (_:_, ... _:_) : _ = t or ? *)
    FnDef of info * (string * ty) list * ty * (term option)

(* top-level commands *)

type command =
    Eval of info * term
  | UnsafeEval of info * term
  | Bind of info * string * binding
  | Prog of info * (string * progdef) list


(* ---------------------------------------------------------------------- *)
(* Contexts *)

type context = (string * binding) list

let emptycontext = []

let isbound ctx x = List.mem_assoc x ctx

let addbinding fi ctx (x, bind) = 
  if isbound ctx x then 
    error fi ("Variable " ^ x ^ " is already bound")
  else (x,bind)::ctx

let getbinding fi ctx x =
  try
    List.assoc x ctx
  with Not_found -> 
    let msg =
      Printf.sprintf "Variable lookup failure: variable %s not found" in
    error fi (msg x)

let getTypeFromContext fi ctx x =
  match getbinding fi ctx x with
    VarBind(tyT) -> tyT

(* ---------------------------------------------------------------------- *)
(* Programs *)

type prog = (string * progdef) list

let emptyprog = []

let isdefined prog fn = List.mem_assoc fn prog

let addprogdef prog (fn,def) = match def with
  FnDef(fi,args,_,bdy_opt) -> 
    if isdefined prog fn then
      error fi ("Function " ^ fn ^ " is already defined")
    else (fn,def)::prog

let makeprog p = List.fold_left addprogdef emptyprog p

let getprogbinding fi prog fn =
  try
    List.assoc fn prog
  with Not_found -> 
    let msg =
      Printf.sprintf "Function name lookup failure: function %s not found" in
    error fi (msg fn)

let getFunBodyFromProgram fi prog fn =
  match getprogbinding fi prog fn with
    FnDef(_,args,_,bdy_opt) -> (List.map fst args,bdy_opt)

let getFunTypeFromProgram fi prog fn =
  match getprogbinding fi prog fn with
    FnDef(_,args,resty,_) -> (List.map snd args,resty)

let hasbodyFun prog fn = 
  match getprogbinding dummyinfo prog fn with
    FnDef(_,_,_,Some _) -> true
  | FnDef(_,_,_,None) -> false

  

(* ---------------------------------------------------------------------- *)
(* Substitution *)

let rec simultSubst subst t = 
  let subit = simultSubst subst in match t with
    TmTrue(_) -> t
  | TmFalse(_) -> t
  | TmIf(fi,t1,t2,t3) -> TmIf(fi,subit t1, subit t2, subit t3)
  | TmZero(_) -> t
  | TmSucc(fi,t) -> TmSucc(fi,subit t)
  | TmPred(fi,t) -> TmPred(fi,subit t)
  | TmIsZero(fi,t) -> TmIsZero(fi,subit t)
  | TmVar(_,x) -> (try (List.assoc x subst) with Not_found -> t)
  | TmFunApp(fi,fn,ts) -> TmFunApp(fi,fn,List.map subit ts)

let simultSubstArgsVals fi args vals bdy =
  try
    simultSubst (List.combine args vals) bdy
  with Invalid_argument _ ->
    let msg =
      Printf.sprintf "Wrong number of arguments for substitution (expected %d, got %d)" in
    error fi (msg (List.length args) (List.length vals))

(* ---------------------------------------------------------------------- *)
(* Extracting file info *)

let tmInfo t = match t with
    TmTrue(fi) -> fi
  | TmFalse(fi) -> fi
  | TmIf(fi,_,_,_) -> fi
  | TmZero(fi) -> fi
  | TmSucc(fi,_) -> fi
  | TmPred(fi,_) -> fi
  | TmIsZero(fi,_) -> fi 
  | TmVar(fi,_) -> fi
  | TmFunApp(fi,_,_) -> fi


(* ---------------------------------------------------------------------- *)
(* Printing *)

(* The printing functions call these utility functions to insert grouping
  information and line-breaking hints for the pretty-printing library:
     obox   Open a "box" whose contents will be indented by two spaces if
            the whole box cannot fit on the current line
     obox0  Same but indent continuation lines to the same column as the
            beginning of the box rather than 2 more columns to the right
     cbox   Close the current box
     break  Insert a breakpoint indicating where the line maybe broken if
            necessary.
  See the documentation for the Format module in the OCaml library for
  more details. 
*)

let obox0() = open_hvbox 0
let obox() = open_hvbox 2
let cbox() = close_box()
let break() = print_break 0 0

let rec printty_Type outer tyT = match tyT with
      tyT -> printty_AType outer tyT

and printty_AType outer tyT = match tyT with
    TyBool -> pr "Bool"
  | TyNat -> pr "Nat"

let printty tyT = printty_Type true tyT 

let rec printtm_Term outer t = match t with
    TmIf(fi, t1, t2, t3) ->
       obox0();
       pr "if ";
       printtm_Term false t1;
       print_space();
       pr "then ";
       printtm_Term false t2;
       print_space();
       pr "else ";
       printtm_Term false t3;
       cbox()
  | t -> printtm_AppTerm outer t

and printtm_AppTerm outer t = match t with
    TmPred(_,t1) ->
       pr "pred "; printtm_ATerm false t1
  | TmIsZero(_,t1) ->
       pr "iszero "; printtm_ATerm false t1
  | TmFunApp(_,fnm,ts) -> 
       pr fnm; pr "(";
       pr_sep "," (printtm_Term outer) ts;
       pr ")"
  | t -> printtm_ATerm outer t

and printtm_ATerm outer t = match t with
    TmTrue(_) -> pr "true"
  | TmFalse(_) -> pr "false"
  | TmZero(fi) ->
       pr "0"
  | TmSucc(_,t1) ->
     let rec f n t = match t with
         TmZero(_) -> pr (string_of_int n)
       | TmSucc(_,s) -> f (n+1) s
       | _ -> (pr "(succ "; printtm_ATerm false t1; pr ")")
     in f 1 t1
  | TmVar(fi,x) -> pr x
  | t -> pr "("; printtm_Term outer t; pr ")"

let printtm t = printtm_Term true t 

let prbinding b = match b with
    VarBind(tyT) -> pr ": "; printty tyT

let prprogdef d = match d with
    FnDef(_,argtys,resty,trm_opt) ->
	pr "(";
        obox0();
        pr_sep "," (fun (x,ty) -> pr x; pr ":"; printty ty; break()) argtys;
        cbox();
        pr "):"; printty resty; pr " = ";
        match trm_opt with
	  None -> pr "?"
        | Some(t) -> break(); printtm t

let prprogbind (f,d) = obox(); pr " def "; pr f; prprogdef d; cbox(); print_newline()
 
let prprog prog = 
    pr "prog"; 
    print_newline();
    List.iter prprogbind prog;
    pr "gorp";
